unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, ComCtrls;

const
  AM_STOP = WM_USER + $96;
  AM_ARRAYMSG = WM_USER + $97;
  AM_ENTRYMSG = WM_USER + $98;
  AM_ERRORMSG = WM_USER + $95;

type
  TMainForm = class(TForm)
    Panel1: TPanel;
    DeviceList: TListBox;
    Panel2: TPanel;
    DetectButton: TButton;
    Panel3: TPanel;
    Panel5: TPanel;
    RepLabel0: TLabel;
    RepLabel1: TLabel;
    RepLabel2: TLabel;
    StatusList: TListBox;
    RepLabel3: TLabel;
    StartInButton: TButton;
    StopButton: TButton;
    RepLabel4: TLabel;
    RepLabel5: TLabel;
    DTCTime: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure DetectButtonClick(Sender: TObject);
    procedure DeviceListClick(Sender: TObject);
    procedure StartButtonClick(Sender: TObject);
    procedure StopButtonClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure DTCTimeTimer(Sender: TObject);
  protected
    procedure AMArrayMsg(var Msg: TMessage); message AM_ARRAYMSG;
    procedure AMEntryMsg(var Msg: TMessage); message AM_ENTRYMSG;
    procedure AMErrorMsg(var Msg: TMessage); message AM_ERRORMSG;
    procedure AMStop(var Msg: TMessage); message AM_STOP;
    { Private declarations }
  public
    { Public declarations }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.DFM}

uses
  AIOUSB;

type
  TReal = Single;

  TStreamThread = class(TThread) //The main worker thread
  protected
    procedure Execute; override;
  public
    hMsgMutex: THandle;
    MsgVal: LongWord;
  end;

  TWriteThread = class(TThread) //Produces blank buffers for the stream thread to take data into
  public
    hInSemaphore, hOutSemaphore,
    hBufMutex: THandle;
    Buf: array[Byte] of Pointer;
    BufIn, BufOut: Byte;
  protected
    procedure Execute; override;
  end;

  TReadThread = class(TThread) //Processes (and frees) buffers
  public
    bSweepData: Boolean;
    hInSemaphore, hOutSemaphore,
    hBufMutex: THandle;
    Buf: array[Byte] of Pointer;
    BufIn, BufOut: Byte;
  protected
    procedure Execute; override;
  end;

var
  DeviceIndex: LongWord;
  StreamThread: TStreamThread;
  WriteThread: TWriteThread;
  ReadThread:  TReadThread;
  TicksPerSec: Int64;
  bFirst: Boolean;
  SmoothedRate: TReal;
  MegatransfersTotal: Integer = 10;
  CheckBlock: Integer;
  BlockSize: Integer = 8192;
  SaveFN: String;





{ TStreamThread }

procedure TStreamThread.Execute;
var
  oTime, nTime: Int64;
  dTime, Ret, L: LongWord;
  Block: Pointer;
  TransfersDone: Integer;

begin
  SetThreadPriority(GetCurrentThread, THREAD_PRIORITY_HIGHEST);

  TransfersDone := 0;
  QueryPerformanceCounter(nTime);
  repeat
    oTime := nTime;

    WaitForSingleObject(WriteThread.hOutSemaphore, INFINITE);
    WaitForSingleObject(WriteThread.hBufMutex, INFINITE);
      Block := WriteThread.Buf[WriteThread.BufOut];
      Inc(WriteThread.BufOut);
    ReleaseMutex(WriteThread.hBufMutex);
    ReleaseSemaphore(WriteThread.hInSemaphore, 1, nil);

    if TransfersDone = 0 then AIOUSB_ClearFIFO(DeviceIndex, 0);
    Ret := DIO_StreamFrame(DeviceIndex, BlockSize, Block, L);

    QueryPerformanceCounter(nTime);

    oTime := nTime - oTime;
    if oTime > High(LongWord) then
      dTime := High(LongWord)
    else
      dTime := oTime
    ;

    L := L div 2;
    Dec(PLongWord(Block));
    PLongWord(Block)^ := L;
    Inc(PLongWord(Block));

    if Ret <> ERROR_SUCCESS then
      PostMessage(MainForm.Handle, AM_ERRORMSG, TransfersDone, Ret)
    else begin
      PostMessage(MainForm.Handle, AM_ARRAYMSG, L, dTime);

      WaitForSingleObject(ReadThread.hBufMutex, INFINITE);
        ReadThread.Buf[ReadThread.BufIn] := Block;
        Inc(ReadThread.BufIn);
      ReleaseMutex(ReadThread.hBufMutex);
      ReleaseSemaphore(ReadThread.hInSemaphore, 1, nil);
    end;

    Inc(TransfersDone);
    if TransfersDone >= (512 + 51) * 1024 then begin
      Terminate;
      PostMessage(MainForm.Handle, AM_STOP, 0, 0);
    end;
  until Terminated;
end;

{ TWriteThread }

procedure TWriteThread.Execute;
var
  Block: PLongWord;

begin
  repeat
    WaitForSingleObject(hInSemaphore, INFINITE);
    if Terminated then Break;

    GetMem(Block, BlockSize * 2 + 4);
    ZeroMemory(Block, BlockSize * 2 + 4);
    Inc(Block);

    WaitForSingleObject(hBufMutex, INFINITE);
      Buf[BufIn] := Block;
      Inc(BufIn);
    ReleaseMutex(hBufMutex);

    ReleaseSemaphore(hOutSemaphore, 1, nil);

  until Terminated;
end;

procedure DumpBufListToStream(const BufList: array of PLongWord; Stream: TStream);
var
  I, L, Ct: Integer;
  pBuf: PLongWord;

begin
  Ct := Length(BufList);
  for I := 0 to Ct - 1 do begin
    PostMessage(MainForm.Handle, AM_ENTRYMSG, I, Ct);

    pBuf := BufList[I];
    Dec(pBuf);
    L := pBuf^;
    Inc(pBuf);

    Stream.Write(pBuf^, L * 2);
  end;
  PostMessage(MainForm.Handle, AM_ENTRYMSG, Ct, Ct);
end;

{ TReadThread }

procedure TReadThread.Execute;
var
  Block: PLongWord;
  I, NextListEntry: Integer;
  OutFil: TFileStream;
  OutList: array of PLongWord;

begin
  NextListEntry := 0;
  if bSweepData then begin
    SetLength(OutList, 512 * 1024);
  end;

  repeat
    WaitForSingleObject(hInSemaphore, INFINITE);
    if Terminated then Break;

    WaitForSingleObject(hBufMutex, INFINITE);
      Block := Buf[BufOut];
      Inc(BufOut);
    ReleaseMutex(hBufMutex);

    if bSweepData then begin
      OutList[NextListEntry] := Block;
      Inc(NextListEntry);
    end
    else begin
      Dec(Block);
      FreeMem(Block, BlockSize * 2 + 4); //Memory management closes the loop, reusing freed buffers of the same size by preference.
    end;

  until Terminated;

  SetLength(OutList, NextListEntry);

  if bSweepData then begin
    OutFil := TFileStream.Create(SaveFN, fmCreate);
    DumpBufListToStream(OutList, OutFil);
    OutFil.Free;
    DIO_Write1(DeviceIndex, 8, True); //Set USB Done high.
    for I := 0 to High(OutList) do begin
      Block := OutList[I];
      Dec(Block);
      FreeMem(Block, BlockSize * 2 + 4);
    end;
    SetLength(OutList, 0);
  end;
end;





function ChunkTo(Len: Integer; S: String): String;
begin
  Result := '';
  while Length(S) > Len do begin
    Result := ' ' + Copy(S, Length(S) - Len + 1, Len) + Result;
    S := Copy(S, 1, Length(S) - Len);
  end;
  Result := S + Result;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  SaveFN := ParamStr(1);
  
  QueryPerformanceFrequency(TicksPerSec);

  DetectButton.Click;

  StartInButton.Click;
end;

procedure TMainForm.DetectButtonClick(Sender: TObject);
var
  DevMask: LongWord;
  I, Sel: Integer;
  NS, PID: LongWord;
  N: String;

begin
  DevMask := GetDevices;

  Sel := -1;

  DeviceList.Items.Clear;
  I := 0;
  while DevMask <> 0 do begin
    if (DevMask and 1) <> 0 then begin
      NS := 256;
      SetLength(N, NS);
      QueryDeviceInfo(I, @PID, @NS, PChar(N), nil, nil);
      SetLength(N, NS);
      DeviceList.Items.Add(IntToStr(I) + ': ' + N + '[' + IntToHex(PID, 4) + ']');
      Sel := 0;
    end;

    DevMask := DevMask shr 1;
    Inc(I);
  end;
  DeviceList.ItemIndex := Sel;
  DeviceList.OnClick(nil);
end;

procedure TMainForm.DeviceListClick(Sender: TObject);
var
  I: Integer;
  Buf: String;

begin
  I := DeviceList.ItemIndex;
  if I < 0 then
    DeviceIndex := $FFFFFFFF
  else begin
    Buf := DeviceList.Items[I];
    I := Pos(':', Buf);
    Buf := Copy(Buf, 1, I - 1);
    DeviceIndex := StrToInt(Buf);
  end;
end;

procedure TMainForm.StartButtonClick(Sender: TObject);
var
  ReadClockHz, WriteClockHz: Double;
  OutMask, TriMask: Byte;
  DIOData: LongWord;

begin
  if DeviceIndex = $FFFFFFFF then begin
    if StartInButton.Visible then
      Application.MessageBox('Please select a board(on the left) first.', 'No Selection', MB_ICONEXCLAMATION)
    else
      StartInButton.Visible := True
    ;
    Exit;
  end;

  SetPriorityClass(GetCurrentProcess, HIGH_PRIORITY_CLASS);

  StatusList.Items.Clear;
  StatusList.ItemIndex := StatusList.Items.Add('Starting...');
  DetectButton.Enabled := False;
  DeviceList.Enabled := False;
  StartInButton.Enabled := False;

  RepLabel0.Caption := '';
  RepLabel1.Caption := '';
  RepLabel2.Caption := '';
  RepLabel3.Caption := '';
  RepLabel4.Caption := '';

  StatusList.ItemIndex := StatusList.Items.Add('Setting Clocks');
  ReadClockHz := 0;
  WriteClockHz := ReadClockHz;
  DIO_StreamSetClocks(DeviceIndex, ReadClockHz, WriteClockHz);
  if ReadClockHz = 0 then
    StatusList.ItemIndex := StatusList.Items.Add('Input Clock: External')
  else
    StatusList.ItemIndex := StatusList.Items.Add('Clocks = ' + IntToStr(Round(ReadClockHz)) + ' Hz');
  ;

  StatusList.ItemIndex := StatusList.Items.Add('StreamOpen');
  DIO_StreamOpen(DeviceIndex, True);
  OutMask := $0A;
  DIOData := $00000000; //Set USB Done (and other outputs) low.
  TriMask := $00;
  DIO_ConfigureEx(DeviceIndex, @OutMask, @DIOData, @TriMask);
  bFirst := True;
  StatusList.ItemIndex := StatusList.Items.Add('Creating Write Thread');
  WriteThread := TWriteThread.Create(True);
  with WriteThread do begin
    FreeOnTerminate := False;
    hInSemaphore := CreateSemaphore(nil, 30, 32, nil);
    hOutSemaphore := CreateSemaphore(nil, 0, 32, nil);
    hBufMutex := CreateMutex(nil, False, nil);
    BufIn := 0;
    BufOut := 0;
    Resume;
  end;
  CheckBlock := 0;
  StatusList.ItemIndex := StatusList.Items.Add('Creating Read Thread');
  ReadThread := TReadThread.Create(True);
  with ReadThread do begin
    FreeOnTerminate := False;
    bSweepData := True;
    hInSemaphore := CreateSemaphore(nil, 0, 512, nil);
    hOutSemaphore := 0;
    hBufMutex := CreateMutex(nil, False, nil);
    BufIn := 0;
    BufOut := 0;
    Resume;
  end;
  StatusList.ItemIndex := StatusList.Items.Add('Creating "Stream" Thread');
  StreamThread := TStreamThread.Create(True);
  with StreamThread do begin
    FreeOnTerminate := False;
    hMsgMutex := CreateMutex(nil, False, nil);
  end;
  StatusList.ItemIndex := StatusList.Items.Add('Waiting for Write Thread to allocate buffers');
  while WaitForSingleObject(WriteThread.hInSemaphore, 0) = WAIT_OBJECT_0 do begin //Wait until the write thread has all buffers ready
    ReleaseSemaphore(WriteThread.hInSemaphore, 1, nil);
    Sleep(1);
  end;
  StatusList.ItemIndex := StatusList.Items.Add('Done Waiting for Write Thread to allocate buffers');
  StatusList.ItemIndex := StatusList.Items.Add('Resuming StreamThread');

  StreamThread.Resume;
  StatusList.ItemIndex := StatusList.Items.Add('Done Resuming StreamThread');

  DTCTime.Enabled := True;
  StatusList.ItemIndex := StatusList.Items.Add('Started DTC Poll');

  StopButton.Enabled := True;
  StatusList.ItemIndex := StatusList.Items.Add('Started.');
end;


procedure TMainForm.StopButtonClick(Sender: TObject);
var
  ReadClockHz, WriteClockHz: Double;
  I: Integer;

begin
  if not StopButton.Enabled then Exit;
  StatusList.ItemIndex := StatusList.Items.Add('Stopping...');
  StopButton.Enabled := False;

  DTCTime.Enabled := False;
  StatusList.ItemIndex := StatusList.Items.Add('Stopped DTC poll...');

  StatusList.ItemIndex := StatusList.Items.Add('Preflushing read thread...');
  StatusList.ItemIndex := StatusList.Items.Add('Terminating read thread...');
  ReadThread.Terminate;
  StatusList.ItemIndex := StatusList.Items.Add('Strobing read thread...');
  for I := 0 to 41 do
    ReleaseSemaphore( ReadThread.hInSemaphore, 1, nil)
  ;
  StatusList.ItemIndex := StatusList.Items.Add('Waiting for read thread to flush...');
  //ReadThread.WaitFor;
  while WaitForSingleObject(ReadThread.Handle, 1) = WAIT_TIMEOUT do Application.ProcessMessages;
  StatusList.ItemIndex := StatusList.Items.Add('Preflushed read thread.');

  StatusList.ItemIndex := StatusList.Items.Add('Stopping stream thread...');
  StatusList.ItemIndex := StatusList.Items.Add('Terminating stream thread...');
  StreamThread.Terminate;
  StatusList.ItemIndex := StatusList.Items.Add('Waiting for stream thread to close...');
  StreamThread.WaitFor;
  StatusList.ItemIndex := StatusList.Items.Add('Freeing stream thread...');
  StreamThread.Free;
  StatusList.ItemIndex := StatusList.Items.Add('Stopping clocks...');
  ReadClockHz := 0;
  WriteClockHz := 0;
  DIO_StreamSetClocks(DeviceIndex, ReadClockHz, WriteClockHz);
  StatusList.ItemIndex := StatusList.Items.Add('Closing stream process...');
  DIO_StreamClose(DeviceIndex);

  StatusList.ItemIndex := StatusList.Items.Add('Stopping write thread...');
  WriteThread.Terminate;
  ReleaseSemaphore(WriteThread.hInSemaphore, 1, nil);
  WriteThread.WaitFor;
  CloseHandle(WriteThread.hInSemaphore);
  CloseHandle(WriteThread.hOutSemaphore);
  CloseHandle(WriteThread.hBufMutex);
  WriteThread.Free;
  StatusList.ItemIndex := StatusList.Items.Add('Stopping read thread...');
  ReadThread.Terminate;
  ReleaseSemaphore( ReadThread.hInSemaphore, 1, nil);
  ReadThread.WaitFor;
  CloseHandle(ReadThread.hInSemaphore);
  CloseHandle(ReadThread.hBufMutex);
  ReadThread.Free;

  DetectButton.Enabled := True;
  DeviceList.Enabled := True;
  StartInButton.Enabled := True;
  StartInButton.Visible := True;
  StatusList.ItemIndex := StatusList.Items.Add('Stopped.');
end;

procedure TMainForm.AMStop(var Msg: TMessage);
begin
  StatusList.Items.Add('Risk of overflow; stopping...');
  StopButton.Click;
end;

procedure TMainForm.AMArrayMsg(var Msg: TMessage);
var
  Amount, TimeTaken: LongWord;

begin
  Amount := Msg.WParam;
  TimeTaken := Msg.LParam;

  if TimeTaken <> $FFFFFFFF then begin
    if bFirst then begin
      bFirst := False;
      SmoothedRate := Amount * TicksPerSec / TimeTaken;
    end
    else begin
      SmoothedRate := SmoothedRate * 0.99 + 0.01 * Amount * TicksPerSec / TimeTaken;
    end;
  end;
end;

procedure TMainForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  if StopButton.Enabled then begin
    case Application.MessageBox('Stop before closing?', 'Stop?', MB_ICONQUESTION or MB_YESNOCANCEL) of
      ID_YES: begin
        StopButton.Click;
        CanClose := True;
      end;
      ID_NO: begin
        CanClose := True;
      end;
      else begin
        CanClose := False;
      end;
    end;
  end;
end;

procedure TMainForm.AMEntryMsg(var Msg: TMessage);
begin
  RepLabel4.Caption := IntToHex(Msg.WParam, 1) + '/' + IntToHex(Msg.LParam, 1);
end;

procedure TMainForm.AMErrorMsg(var Msg: TMessage);
begin
  RepLabel3.Caption := 'Error #' + IntToStr(Msg.LParam) + ' at ' + IntToHex(Msg.WParam, 1);
end;

procedure TMainForm.DTCTimeTimer(Sender: TObject);
var
  DTC: Byte;

begin
  RepLabel0.Caption := 'Throughput: ' + ChunkTo(3, IntToStr(Round(SmoothedRate))) + ' T/s';

  DIO_Read1(DeviceIndex, 0, @DTC);
  if DTC <> 0 then begin
    StreamThread.Terminate;
    StatusList.Items.Add('Data transmission complete, stopping...');

    StopButton.Click;
  end;
end;

end.
