// StreamSaver.cpp : Defines the entry point for the console application.
//

//#define BENCHMARK_READ
//#define BENCHMARK_WRITE

#include "stdafx.h"
#include <stdlib.h>
#include <conio.h>
#include <iostream.h>
#include <windows.h>
#include "AIOUSB.h"

const unsigned long FramePoints = 8*1024;
LARGE_INTEGER TicksPerSec;

typedef struct _TAcqBlock
{
	LARGE_INTEGER AcqStartTime, AcqBackTime;
	unsigned long Len;
	struct _TAcqBlock *NextBlock; //One-way linked list, used as a FIFO.
	unsigned short Block[FramePoints];
} TAcqBlock, *pTAcqBlock;

class TAcqList
{
	public:
		TAcqList(void);
		~TAcqList(void);

	protected:
		pTAcqBlock FirstBlock, LastBlock;
		HANDLE hListMutex, hListSem;

	public:
		virtual void UseSem(void);
		virtual void EndBlocks(void);
		virtual void WriteBlock(pTAcqBlock NewBlock);
		virtual pTAcqBlock ReadBlock(void);
};

class TWorkerThread
{
	public:
		TWorkerThread(void);
		~TWorkerThread(void);

	protected:
		int bTerminated, bBegun;
		HANDLE hThread;
		int MyPriority;

	public:
		unsigned long ReturnValue;

		virtual int BeginThread(void);
		virtual void Terminate(void);
		virtual unsigned long ThreadProc(void) { return ERROR_SUCCESS; };
		int IsDone(void);
		void WaitFor(void);
};
typedef TWorkerThread *pTWorkerThread;

class TFileThread : public TWorkerThread
{
	public:
		TFileThread(void);
		~TFileThread(void);

	public:
		HANDLE hFile;
		TAcqList InList;
		double Throughput;

		unsigned long ThreadProc(void);
} *FileThread;

class TDI16AThread : public TWorkerThread
{
	public:
		unsigned long DeviceIndex;

		unsigned long ThreadProc(void);
} *DI16AThreadA, *DI16AThreadB;

class TNewThread : public TWorkerThread
{
	public:
		TNewThread(void);
		~TNewThread(void);

	public:
		HANDLE hPrebufferSem;
		TAcqList InList, OutList;

		unsigned long ThreadProc(void);
} *NewThread;



char * IntToHexStatic(int I)
{
	static char Buf[10];
	itoa(I, Buf, 16);
	return Buf;
}

void PressAnyKeyToClose(void)
{
	cout << "Press any key to close." << endl;
	getch();
}

int main(int argc, char* argv[])
{
	unsigned long Status, PID;
	unsigned long DeviceIndex = diFirst;
	HANDLE hFile;
	int bGoing;
	unsigned char bDTC = 0;
	char *SaveFile;
	signed long FilePosHigh;

	if ( argc < 2 )
	{
		cout << "Usage: StreamSaver filename" << endl;
		cout << endl;
		cout << "The file specified by name will be created (it must not exist yet)," << endl;
		cout << "and data will be saved to it." << endl;
		PressAnyKeyToClose();
		return -1;
	}
	SaveFile = argv[1];

	Status = QueryDeviceInfo(DeviceIndex, &PID, NULL, NULL, NULL, NULL);
	if ( Status != ERROR_SUCCESS )
	{
		cout << "Error " << Status << " querying USB board." << endl;
		PressAnyKeyToClose();
		return -2;
	}
	if ( (PID & 0xFFF8) != 0x8008 )
	{
		cout << "USB board detected(PID " << IntToHexStatic(PID) << ") was not USB-DIO16A series." << endl;
		PressAnyKeyToClose();
		return -3;
	}

	hFile = CreateFile(SaveFile, GENERIC_WRITE, 0, NULL, CREATE_NEW, FILE_FLAG_SEQUENTIAL_SCAN | FILE_ATTRIBUTE_NORMAL, 0);
	if ( hFile == INVALID_HANDLE_VALUE )
	{
		Status = GetLastError();
		cout << "Error ";
		if ( Status == ERROR_FILE_EXISTS )
			cout << "ERROR_FILE_EXISTS";
		else
			cout << Status
		;
		cout << " creating file \"" << SaveFile << "\" for write." << endl;
		PressAnyKeyToClose();
		return -4;
	}
	//Pad the file out to 2GB.
	FilePosHigh = 0;
	Status = SetFilePointer(hFile, 0x80000000, &FilePosHigh, FILE_BEGIN);
	SetEndOfFile(hFile);
	SetFilePointer(hFile, 0, &FilePosHigh, FILE_BEGIN);

	QueryPerformanceFrequency(&TicksPerSec);

	FileThread = new TFileThread;
	FileThread->hFile = hFile;
	FileThread->BeginThread();

	NewThread = new TNewThread;
	ReleaseSemaphore(NewThread->hPrebufferSem, 16, NULL);
	NewThread->BeginThread();
	Sleep(1); //Give the new() thread time to prebuffer some blocks.

	{
		__int8 OutMask = 0xA; //Ports B and D are output; A0 is Data Transmission Complete, B0 is USB Done, D and C must be opposite.
		unsigned long Data = 0x00000000; //Set USB Done (and all other outputs) low.
		__int8 TristateMask = 0x00; //Untristate all.
		double ReadClockHz = 0, WriteClockHz = 0;

		DIO_ConfigureEx(DeviceIndex, &OutMask, &Data, &TristateMask);
		DIO_StreamSetClocks(DeviceIndex, &ReadClockHz, &WriteClockHz);
		DIO_StreamOpen(DeviceIndex, true);
	}

	DI16AThreadA = new TDI16AThread;
	DI16AThreadA->DeviceIndex = DeviceIndex;
	DI16AThreadA->BeginThread();

	DI16AThreadB = new TDI16AThread;
	DI16AThreadB->DeviceIndex = DeviceIndex;
	DI16AThreadB->BeginThread();

	Status = ERROR_SUCCESS;
	bGoing = true;
	while ( bGoing )
	{
		Sleep(1); //Give other threads (and processes) a turn.

		Status = ERROR_SUCCESS;
		if ( DI16AThreadA->IsDone() ) Status = DI16AThreadA->ReturnValue;
		if ( DI16AThreadB->IsDone() ) Status = DI16AThreadB->ReturnValue;
		//The DIO worker threads will only exit on their own in case of error.

		if ( Status == ERROR_SUCCESS )
		{
			DIO_Read1(DeviceIndex, 0, &bDTC);
			if ( bDTC )
			{
				cout << "Data transmission is complete. Ending process..." << endl;
				bGoing = false;
			}
		}
		else
		{
			cout << "Error " << Status << " streaming a frame. Aborting process..." << endl;
			bGoing = false;
		}
	}

	DIO_StreamClose(DeviceIndex);

	//Deleting the DIO worker threads waits for them to finish flushing to the
	//file thread.
	delete DI16AThreadA;
	delete DI16AThreadB;

	NewThread->Terminate();
	NewThread->InList.EndBlocks();

	FileThread->InList.EndBlocks();
	FileThread->WaitFor();

	DIO_Write1(DeviceIndex, 8, true); //Set USB Done high.

	cout << "Average rate: " << long(FileThread->Throughput) << endl;

	delete FileThread;
	delete NewThread;

	SetEndOfFile(hFile);
	CloseHandle(hFile);

	PressAnyKeyToClose();

	return Status;
}



TAcqList::TAcqList(void)
{
	FirstBlock = NULL;
	LastBlock = NULL;
	hListSem = 0;
	hListMutex = CreateMutex(NULL, false, NULL);
}

TAcqList::~TAcqList(void)
{
	pTAcqBlock ThisBlock, NextBlock;

	WaitForSingleObject(hListMutex, INFINITE);
		ThisBlock = FirstBlock;
		FirstBlock = NULL;
		LastBlock = NULL;
	ReleaseMutex(hListMutex);

	while ( ThisBlock )
	{
		NextBlock = ThisBlock->NextBlock;
		delete ThisBlock;
		ThisBlock = NextBlock;
	}

	CloseHandle(hListMutex);
	if ( hListSem ) CloseHandle(hListSem);
}

void TAcqList::WriteBlock(pTAcqBlock NewBlock)
{
	NewBlock->NextBlock = NULL;

	WaitForSingleObject(hListMutex, INFINITE);
		if ( LastBlock )
			LastBlock->NextBlock = NewBlock;
		else
			FirstBlock = NewBlock
		;
		LastBlock = NewBlock;
	ReleaseMutex(hListMutex);

	if ( hListSem ) ReleaseSemaphore(hListSem, 1, NULL);
}

pTAcqBlock TAcqList::ReadBlock(void)
{
	pTAcqBlock ThisBlock = NULL;

	if ( hListSem ) WaitForSingleObject(hListSem, INFINITE);

	WaitForSingleObject(hListMutex, INFINITE);
		if ( FirstBlock )
		{
			ThisBlock = FirstBlock;
			if ( FirstBlock->NextBlock )
			{
				FirstBlock = FirstBlock->NextBlock;
			}
			else
			{
				FirstBlock = NULL;
				LastBlock = NULL;
			}
		}
	ReleaseMutex(hListMutex);

	return ThisBlock;
}

void TAcqList::EndBlocks(void)
{
	ReleaseSemaphore(hListSem, 1, NULL);
}

void TAcqList::UseSem(void)
{
	hListSem = CreateSemaphore(NULL, 0, 0x40000000, NULL);
}



unsigned long WINAPI WorkerThreadRoot(LPVOID pParam)
{
	TWorkerThread *Self = pTWorkerThread(pParam);
	Self->ReturnValue = Self->ThreadProc();
	return Self->ReturnValue;
}

TWorkerThread::TWorkerThread(void)
{
	bTerminated = false;
	bBegun = false;
	ReturnValue = ERROR_IO_PENDING;
	//Boost worker threads relative to main thread, because they spend most of
	//their time locked up waiting for I/O.
	MyPriority = THREAD_PRIORITY_HIGHEST;
	hThread = INVALID_HANDLE_VALUE;
}

TWorkerThread::~TWorkerThread(void)
{
	bTerminated = true;
	
	if ( hThread && (hThread != INVALID_HANDLE_VALUE) && bBegun )
	{
		WaitForSingleObject(hThread, INFINITE);
	}
}

int TWorkerThread::BeginThread(void)
{
	unsigned long TID;

	if ( bBegun ) return false;

	bBegun = true;
	hThread = CreateThread(NULL, 0, &WorkerThreadRoot, this, 0, &TID);
	if ( hThread == NULL )
	{
		bBegun = false;
	}
	else
	{
		SetThreadPriority(hThread, MyPriority);
	}

	return bBegun;
}

void TWorkerThread::Terminate(void)
{
	bTerminated = true;
}

void TWorkerThread::WaitFor(void)
{
	WaitForSingleObject(hThread, INFINITE);
}

int TWorkerThread::IsDone(void)
{
	return WaitForSingleObject(hThread, 0) == WAIT_OBJECT_0;
}



TFileThread::TFileThread(void)
{
	Throughput = 0;
	MyPriority = THREAD_PRIORITY_ABOVE_NORMAL;
	InList.UseSem();
}

TFileThread::~TFileThread(void)
{
	InList.EndBlocks();
}

unsigned long TFileThread::ThreadProc(void)
{
	signed long Amount = -1;
	__int64 TimeTaken = 0;

	while ( ! bTerminated )
	{
		pTAcqBlock ThisBlock;
		ThisBlock = InList.ReadBlock();
		if ( ThisBlock )
		{
			if ( Amount == -1 )
			{
				//Skip first transfer, because its time includes the waiting for RX_EN to activate.
				Amount = 0;
			}
			else
			{
				Amount += ThisBlock->Len;
				TimeTaken += ThisBlock->AcqBackTime.QuadPart - ThisBlock->AcqStartTime.QuadPart;
			}

			while ( ThisBlock->Len )
			{
				unsigned long BytesTransferred = 0;
				WriteFile(hFile, ThisBlock->Block, ThisBlock->Len, &BytesTransferred, NULL);
				ThisBlock->Len -= BytesTransferred;
			}
			NewThread->InList.WriteBlock(ThisBlock);
		}
		else
		{
			bTerminated = true;
		}
	}

	Throughput = double(Amount) * double(TicksPerSec.QuadPart) / double(TimeTaken);

	return ERROR_SUCCESS;
}



unsigned long TDI16AThread::ThreadProc(void)
{
	unsigned long Status = ERROR_SUCCESS;

	while ( ! bTerminated )
	{
		pTAcqBlock ThisBlock = NewThread->OutList.ReadBlock();

		if ( ThisBlock == NULL )
		{
			bTerminated = true;
		}
		else
		{
			QueryPerformanceCounter(&ThisBlock->AcqStartTime);
			Status = DIO_StreamFrame(DeviceIndex, FramePoints, ThisBlock->Block, &ThisBlock->Len);
			QueryPerformanceCounter(&ThisBlock->AcqBackTime);

			if ( Status == ERROR_SUCCESS )
			{
				FileThread->InList.WriteBlock(ThisBlock);
				//Since DIO_StreamFrame() won't transfer while some other thread
				//is transferring, a race condition is prevented for successful
				//transfers.
			}
			else
			{
				bTerminated = true;
			}
		}
	}

	return Status;
}



TNewThread::TNewThread(void)
{
	MyPriority = THREAD_PRIORITY_ABOVE_NORMAL;
	hPrebufferSem = CreateSemaphore(NULL, 0, 0x40000000, NULL);
	InList.UseSem();
	OutList.UseSem();
}

TNewThread::~TNewThread(void)
{
	CloseHandle(hPrebufferSem);
	InList.EndBlocks();
}

unsigned long TNewThread::ThreadProc(void)
{
	pTAcqBlock ThisBlock;

	while ( (WaitForSingleObject(hPrebufferSem, 0) == WAIT_OBJECT_0) )
	{
		OutList.WriteBlock(new TAcqBlock);
	}

	while ( ! bTerminated )
	{
		ThisBlock = InList.ReadBlock();
		if ( ThisBlock )
		{
			OutList.WriteBlock(ThisBlock);
		}
		else
		{
			bTerminated = true;
		}
	}

	return ERROR_SUCCESS;
}

